/******************************************************************************
*
* project name:    TICHESS
* project version: 1.00
* file name:       engine.c
* file version:    1.03
* initial date:    24/02/2000
* author:          thomas.nussbaumer@peem.com
* description:     contents the chess playing engine's stuff
*
*
* [based on ideas and code of minimax engine from Ch.Donninger]
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.00: 24/02/2000
*        extracted from tichess.c
*
* v1.01  27/02/2000
*        statistics printing after each computer move (compiler define)
*        
* v1.02  28/02/2000
*        cancelation of calculation will query user
*
* v1.03  29/02/2000
*        engine speedup by updating pawn variables "on-the-fly"
*
******************************************************************************/

#include <nostub.h>
#include <all.h>
#include "tichess.h"

/*===========================================================================*/
/*===========================================================================*/
/*===========================================================================*/
/*                                                                           */
/* G L O B A L S                                                             */
/*                                                                           */
/*===========================================================================*/
/*===========================================================================*/
/*===========================================================================*/

char       txt_rlong[]           = "O-O-O"; // notation for long rochade
char       txt_rshort[]          = "O-O";   // notation for short rochade
char       movetext[8]           = {'0'};   // used by Move2Str() routine
char       updatecounter         = 0;       // used to overcome LONG modulo
                                            // operation (size!!)
int        already_matt          = FALSE;
char*      board                 = 0;  // the board
int*       ep_field              = 0;
move_t*    move_stack            = 0;
int*       move_indicators       = 0;  // counts figures which leaves field
int*       stack_ranges          = 0;
fromto_t** main_var              = 0;  // table of main variants
killer_t*  killer_tab            = 0;  // table of killer moves
bicolor_t* p_controlled          = 0;  // fields controlled by pawns
bicolor_t* p_per_line            = 0;  // number of pawns on a line
bicolor_t* r_per_line            = 0;  // number of rooks on a line


int*       mobility              = 0;  // mobility of bishop and rook
int*       to_fields             = 0;  // used for sorting and search ext.
int*       material_bal          = 0;  // material balance  white/black
int*       material_sum          = 0;  // material sum
int        rochade_indicators[3] = {}; // indicates white/black rochade
int        act_index             = 0;  // actual index in move_stack
int        white_king_pos        = 0;  // position of white king
int        black_king_pos        = 0;  // position of black king
int        act_color             = 0;  // which turn (white or black)
int        std_depth             = 0;  // standard depth
int        max_extend            = 0;  // max.extention of search tree
int        act_depth             = 0;  // actual search depth
long       nr_nodes              = 0;  // number of evaluated nodes
int        last_move             = 0;  // last executed move
int        chess_indicator       = 0;  // actual side is in chess
int        move_count            = 0;  // number of moves so far


/*---------------------------------------------------------------------------*/
/* initial board setup                                                       */
/*---------------------------------------------------------------------------*/

const char board_setup[BOARD_SIZE] =  { 
OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,  
OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE, 
OUTSIDE,W_ROOK, W_KNIGHT,W_BISHOP,W_QUEEN,W_KING, W_BISHOP,W_KNIGHT,W_ROOK, OUTSIDE,
OUTSIDE,W_PAWN, W_PAWN,  W_PAWN,  W_PAWN, W_PAWN, W_PAWN,  W_PAWN,  W_PAWN, OUTSIDE,
OUTSIDE,EMPTY,  EMPTY,   EMPTY,   EMPTY,  EMPTY,  EMPTY,   EMPTY,   EMPTY,  OUTSIDE,
OUTSIDE,EMPTY,  EMPTY,   EMPTY,   EMPTY,  EMPTY,  EMPTY,   EMPTY,   EMPTY,  OUTSIDE,
OUTSIDE,EMPTY,  EMPTY,   EMPTY,   EMPTY,  EMPTY,  EMPTY,   EMPTY,   EMPTY,  OUTSIDE,
OUTSIDE,EMPTY,  EMPTY,   EMPTY,   EMPTY,  EMPTY,  EMPTY,   EMPTY,   EMPTY,  OUTSIDE,
OUTSIDE,B_PAWN, B_PAWN,  B_PAWN,  B_PAWN, B_PAWN, B_PAWN,  B_PAWN,  B_PAWN, OUTSIDE,
OUTSIDE,B_ROOK, B_KNIGHT,B_BISHOP,B_QUEEN,B_KING, B_BISHOP,B_KNIGHT,B_ROOK, OUTSIDE,
OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,  
OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE,OUTSIDE, OUTSIDE, OUTSIDE,OUTSIDE}; 

/*---------------------------------------------------------------------------*/
/* move generator tables                                                     */
/*---------------------------------------------------------------------------*/
const char offset_vals[NR_DIRS] = { 
   -9,-11,9,11,                  /* bishop directions  */
   -1,10,1,-10,                  /* rook               */
   19,21,12,-8,-19,-21,-12,8};   /* knight             */

const figoffset_t figure_offsets[NR_FIGS] = {
   {0,0,FALSE},   /* empty field                         */
   {-1,-1,FALSE}, /* pawn moves are generated separately */
   {4,7,TRUE},    /* rook                                */
   {0,3,TRUE},    /* bishop                              */
   {8,15,FALSE},  /* knight                              */
   {0,7,TRUE},    /* queen                               */
   {0,7,FALSE}};  /* king                                */

/*---------------------------------------------------------------------------*/
/* table used for king position evaluation                                   */
/*---------------------------------------------------------------------------*/
const char center_table[H8+1] = {
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,         
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,         
   0,  4,  0,  8, 12, 12,  8,  0,  4,  0,   
   0,  4,  8, 12, 16, 16, 12,  8,  4,  0,   
   0,  8, 12, 16, 20, 20, 16, 12,  8,  0,   
   0, 12, 16, 20, 24, 24, 20, 16, 12,  0,   
   0, 12, 16, 20, 24, 24, 20, 16, 12,  0,   
   0,  8, 12, 16, 20, 20, 16, 12,  8,  0,   
   0,  4,  8, 12, 16, 16, 12,  8,  4,  0,   
   0,  4,  0,  8, 12, 12,  8,  0,  4};      

/*---------------------------------------------------------------------------*/
/* field values for white pawns                                              */
/*---------------------------------------------------------------------------*/
const char wp_fieldvalues[H7+1]= {   
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,         
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,         
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,  
   0,  4,  4,  0,  0,  0,  6,  6,  6, 0,  
   0,  6,  6,  8,  8,  8,  4,  6,  6, 0,   
   0,  8,  8, 16, 22, 22,  4,  4,  4, 0,   
   0, 10, 10, 20, 26, 26, 10, 10, 10, 0,   
   0, 12, 12, 22, 28, 28, 14, 14, 14, 0,   
   0, 18, 18, 28, 32, 32, 20, 20, 20};     

/*---------------------------------------------------------------------------*/
/* field values for black pawns                                              */
/*---------------------------------------------------------------------------*/
const char bp_fieldvalues[H7+1]= {    
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,         
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,         
   0,  0,  0,  0,  0,  0,  0,  0,  0, 0,  
   0, 18, 18, 28, 32, 32, 20, 20, 20, 0,  
   0, 12, 12, 22, 28, 28, 14, 14, 14, 0,  
   0, 10, 10, 20, 26, 26, 10, 10, 10, 0,  
   0,  8,  8, 16, 22, 22,  4,  4,  4, 0,  
   0,  6,  6,  8,  8,  8,  4,  6,  6, 0,  
   0,  4,  4,  0,  0,  0,  6,  6,  6};    

/*---------------------------------------------------------------------------*/
/* figure values                                                             */
/*---------------------------------------------------------------------------*/
const int fig_values[NR_FIGS]= {0,VALUE_PAWN,VALUE_ROOK,VALUE_BISHOP,
                                VALUE_KNIGHT,VALUE_QUEEN,VALUE_KING};
/*---------------------------------------------------------------------------*/
/* figure symbols                                                            */
/*---------------------------------------------------------------------------*/
const char fig_symbols[NR_FIGS] = { '.','P','R','B','N','Q','K' };



#if DO_STATISTICS

long attackcount = 0;
long nodescount  = 0;
long bpevalcount = 0;
long wpevalcount = 0;

int  showstat = FALSE;

void ToggleShowStats() {
    if (showstat == TRUE) {
        puts("statistics printing OFF\n");
        showstat = FALSE;
    }
    else {
        puts("statistics printing ON\n");
        showstat = TRUE;
    }
}

void InitStatistics() {
   showstat    = FALSE;
   attackcount = 0;
   wpevalcount = 0;
   bpevalcount = 0;
   nodescount  = 0;
}

void ShowStatistics() {
    if (showstat == TRUE) {
        printf("attacksField()=%ld\n",attackcount);
        printf("WhitePawnEval()=%ld\n",wpevalcount);
        printf("BlackPawnEval()=%ld\n",bpevalcount);
        printf("nodes sum=%ld\n",nodescount);
    }
}

#endif





#if CHECK_PAWN_STRUCT
void CheckPawnVars(void);
#endif



void  CopyMainVariant(int idx);
void  ExecMove(int idx);
void  UndoMove(int idx);
void  InitPawnVars(void);
void  GenerateMoves(int allmoves);

//-----------------------------------------------------------------------------
// declarations of the necessary "external" routines and variables
//-----------------------------------------------------------------------------
void  UpdateMoveDisplay(move_t* pmove);
void  UpdateNodeDisplay(long nodes);
void  LeavePrg(char* msg);
void  puts(const char *s);



/*===========================================================================*/
/* allocates memory (returns true if everything is allocated)                */
/*===========================================================================*/
int InitMemory() {
    int i,j;
    if ((board      = (char*)malloc(sizeof(char)*BOARD_SIZE))==0)         return FALSE;
    if ((ep_field   = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)            return FALSE;
    if ((move_stack = (move_t*)malloc(sizeof(move_t)*(STACK_SIZE+1)))==0) return FALSE;
    if ((move_indicators = (int*)malloc(sizeof(int)*(H8+1)))==0)          return FALSE;
    if ((stack_ranges  = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)         return FALSE;
    if ((killer_tab    = (killer_t*)malloc(sizeof(killer_t)*MAX_DEPTH))==0)      return FALSE;
    if ((p_controlled  = (bicolor_t*)malloc(sizeof(bicolor_t)*BOARD_SIZE))==0)   return FALSE;
    if ((p_per_line    = (bicolor_t*)malloc(sizeof(bicolor_t)*(H_COLUMN+2)))==0) return FALSE;
    if ((r_per_line    = (bicolor_t*)malloc(sizeof(bicolor_t)*(H_COLUMN+2)))==0) return FALSE;
    if ((mobility      = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)  return FALSE;
    if ((to_fields     = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)  return FALSE;
    if ((material_bal  = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)  return FALSE;
    if ((material_sum  = (int*)malloc(sizeof(int)*MAX_DEPTH))==0)  return FALSE;

    // special treatment for main_var because it is an array of dimension 2
    if ((main_var = (fromto_t**)malloc(sizeof(fromto_t*)*MAX_DEPTH))==0) return FALSE;
    for (i=0;i<MAX_DEPTH;i++) {
        main_var[i] = (fromto_t*)malloc(sizeof(fromto_t)*MAX_DEPTH);
        if (main_var[i] == 0) {
            for (j=0;j<i;j++) {
                free(main_var[i]);
                main_var[i] = 0;
            }
            return FALSE;
        }
    }

    return TRUE;
}


/*===========================================================================*/
/* frees previously allocated memory and sets pointers to zero               */
/*===========================================================================*/
void FreeMemory() {
    int i;
    if (board != 0)           {free(board);board=0;}
    if (ep_field != 0)        {free(ep_field);ep_field=0;}
    if (move_stack != 0)      {free(move_stack);move_stack=0;}
    if (move_indicators != 0) {free(move_indicators);move_indicators=0;}
    if (stack_ranges != 0)    {free(stack_ranges);stack_ranges=0;}
    if (killer_tab != 0)      {free(killer_tab);killer_tab=0;}
    if (p_controlled != 0)    {free(p_controlled);p_controlled=0;}
    if (p_per_line != 0)      {free(p_per_line);p_per_line=0;}
    if (r_per_line != 0)      {free(r_per_line);r_per_line=0;}
    if (mobility != 0)        {free(mobility);mobility=0;}
    if (to_fields != 0)       {free(to_fields);to_fields=0;}
    if (material_bal != 0)    {free(material_bal);material_bal=0;}
    if (material_sum != 0)    {free(material_sum);material_sum=0;}
    if (main_var != 0) {
        for (i=0;i<MAX_DEPTH;i++) {
            if (main_var[i] != 0) {
                free(main_var[i]);
                main_var[i] = 0;
            }
            else {
                break;
            }
        }
        free(main_var);
        main_var = 0;
    }

}




/*===========================================================================*/
/* Initialization of board and game state                                    */
/*===========================================================================*/
void InitEngine(int level) {
    int i;

    already_matt   = FALSE;
    updatecounter  = 0;    

    //------------------------------------------------
    // reset move counter and copy initial board setup
    //------------------------------------------------
    move_count = 0;
    for(i=0;i<BOARD_SIZE;i++) {
       board[i] = board_setup[i];
    }

    //------------------------------------------------
    // reset positions of kings
    // reset enpassant state
    // reset rochade markers
    // reset move indicators
    //------------------------------------------------
    white_king_pos = E1;
    black_king_pos = E8;
    ep_field[0]    = ILLEGAL;

    for (i=0;i<3;i++)   rochade_indicators[i] = FALSE;
    for(i=A1;i<=H8;i++) move_indicators[i]=0;

    //------------------------------------------------
    // reset material sum
    // reset material balance
    // reset first stack range
    // set standard depth
    // reset actual depth
    // set actual color is WHITE
    //------------------------------------------------
    material_sum[0] = VALUE_MAX;
    material_bal[0] = 0;
    stack_ranges[0] = 0;
    std_depth       = level;
    act_depth       = 0;
    act_color       = WHITE;
}


/*===========================================================================*/
/* mapping of figure value to character                                      */
/*===========================================================================*/
char Figure2Char(char figure) {
    return fig_symbols[abs(figure)];
}


/*===========================================================================*/
/* converts internal field number into field notation ("--" == invalid field)*/
/*===========================================================================*/
char* Field2Notation(int fieldnumber,char* fieldnotation) {
    //---------------------
    // validate given field
    //---------------------
    if (fieldnumber<A1 || fieldnumber>H8 || board[fieldnumber]==OUTSIDE) {
        strcpy(fieldnotation,"--");
    }
    else {
        fieldnotation[0] = 'A'- 1 + fieldnumber%10;
        fieldnotation[1] = '1'- 2 + (fieldnumber/10);
        fieldnotation[2] = 0;
    }
    return fieldnotation;
}


/*===========================================================================*/
/* converts field notation into internal field number                        */
/*===========================================================================*/
int Notation2Field(char* fieldnotation) {
    //-----------------------------
    // validate given fieldnotation
    //-----------------------------
    if (fieldnotation[0]<'A' ||
        fieldnotation[0]>'H' ||
        fieldnotation[1]<'1' ||
        fieldnotation[1]>'8' )
    {
       return ILLEGAL;
    }
    return (fieldnotation[0]-'A'+1) + 10*(fieldnotation[1]-'1'+2);
}


/*===========================================================================*/
/* converts move to string representation                                    */
/*===========================================================================*/
char* Move2Str(move_t *pmove) {
    int from,to;

    if (pmove->rochade_nr==ROCHADE_LONG) {
        return txt_rlong;
    }
    if (pmove->rochade_nr==ROCHADE_SHORT) {
        return txt_rshort;
    }

    from = pmove->from;
    to   = pmove->to;

    movetext[0]=fig_symbols[abs(board[from])];

    Field2Notation(from,&movetext[1]);

    if (pmove->killed_fig==EMPTY) {
        movetext[3] = '-';
    }
    else {
        movetext[3] = 'x';
    }

    Field2Notation(to,&movetext[4]);

    if (pmove->promoted_fig!=EMPTY) {
        movetext[6] = fig_symbols[pmove->promoted_fig];
        movetext[7] = 0;
    }
    else {
        movetext[6] = 0;
    }

    return movetext;
}


/*===========================================================================*/
/* intialize tree                                                            */
/*===========================================================================*/
void InitTree(void) {
    if(act_depth==0) return; // already initialized

    ep_field[0]     = ep_field[1];
    material_bal[0] = material_bal[1];
    material_sum[0] = material_sum[1];
    act_depth       = 0;
}


/*===========================================================================*/
/* returns next move not played until now (returns -1 if all are played)     */
/*===========================================================================*/
int NextBestMove(void) {
    int i,bestval,bestmove;
    bestmove = -1;
    bestval  = -VALUE_MATT;

    //------------------------------------------------------------------------
    // loop through all moves of act_depth (range is stored in stack_ranges[])
    //------------------------------------------------------------------------
    for (i=stack_ranges[act_depth];i<stack_ranges[act_depth+1];i++) {
        if (move_stack[i].value > bestval) {
            bestmove = i;
            bestval  = move_stack[i].value;
        }
    }

    //------------------------------------------------------------------------
    // if we've found a best move, set its value to -VALUE_MATT.
    // This will prevent us from selecting the same move later.
    //------------------------------------------------------------------------
    if(bestmove >= 0) move_stack[bestmove].value=-VALUE_MATT;

    return bestmove;
}


/*===========================================================================*/
/* alpha/beta tree search: returns evaluation from actual player's sight     */
/*                                                                           */
/* alpha    ... lower border                                                 */
/* beta     ... upper border                                                 */
/* distance ... number of moves to the horizon                               */
/*                                                                           */
/* if distance is positive --> normal alpha beta searching                   */
/* otherwise               --> "Ruhesuche"                                   */
/*                                                                           */
/* returns NegaMax value from actual player's sight                          */
/*                                                                           */
/* (WARNING: recursive function)                                             */
/*===========================================================================*/
int AlphaBetaSearch(int alpha,int beta,int distance) {
    int i,val,bestvalue,chess;

    nr_nodes++;       // increment node counter
    updatecounter++;
    if (updatecounter==100) {
        UpdateNodeDisplay(nr_nodes);
        updatecounter=0;
    }

    main_var[act_depth][act_depth].from=0;   // deletes act. main variant

    val = Evaluate(alpha,beta,act_color);

    if ((chess=chess_indicator)==TRUE && act_depth+distance<max_extend+1) {
        distance++;  // extend search
    }
    else if ( act_depth >= 2                                 &&
              act_depth+distance < max_extend                &&
              to_fields[act_depth] == to_fields[act_depth-1] &&
              val >= alpha-150                               &&
              val <= beta+150)
    {
        distance++;  // extend search
    }

    if (distance<-5 || val==VALUE_MATT-act_depth || act_depth>=MAX_DEPTH) {
        return val;
    }

    if (val>=beta && distance+chess<=1) {
        return val;
    }

    GenerateMoves(distance);

    bestvalue = (distance>0) ? -VALUE_MATT : val;


    //----------------------------------
    // examine all moves in sorted order
    //----------------------------------
    while ((i=NextBestMove())>=0) {
        ExecMove(i);
        //----------------------------------------------------
        // NegaMax prinziple: change sign, swap alpha and beta
        //----------------------------------------------------
        val = -AlphaBetaSearch(-beta,-alpha,distance-1);
        UndoMove(i);

        if (val>bestvalue) {
            bestvalue=val;
            if (val>=beta) {
                if (distance>0) {
                    CopyMainVariant(i);
                }
                goto done;
            }
            if (val>alpha) {
                if (distance>0) {
                    CopyMainVariant(i);
                }
                alpha=val;
            }
        }
    }

done:
    //------------------------------------------------------------
    // good move, generate cutoff, will be entered in killer table
    // store previous killer move as second best killer
    //------------------------------------------------------------
    if (val>=beta && i>=0) {
        killer_tab[act_depth].killer2=killer_tab[act_depth].killer1;
        killer_tab[act_depth].killer1.from=move_stack[i].from;
        killer_tab[act_depth].killer1.to=move_stack[i].to;
    }

    if (bestvalue == -(VALUE_MATT-(act_depth+1)) && chess == FALSE) return 0;

    return bestvalue;
}


/*===========================================================================*/
/* stores actual move into main variant and copies continuation on the       */
/* next depth                                                                */
/*===========================================================================*/
void CopyMainVariant(int idx) {
    int i;
    /* new main variant will continue this variant */
    main_var[act_depth][act_depth].from=move_stack[idx].from;
    main_var[act_depth][act_depth].to=move_stack[idx].to;
    for (i=1;;i++) {
        if (main_var[act_depth+1][act_depth+i].from==0) {
            main_var[act_depth][act_depth+i]=main_var[act_depth+1][act_depth+i];
            return;
        }
        main_var[act_depth][act_depth+i]=main_var[act_depth+1][act_depth+i];
    }
}


/*===========================================================================*/
/* executes a move on the board and updates state and search depth           */
/*===========================================================================*/
void ExecMove(int idx) {
    int from,to,epfield,matchange,promoted,tmp_to,figure;

    // increment move counter
    move_count++;

    // setup locals and update last_move global
    from      = move_stack[idx].from;
    to        = move_stack[idx].to;
    epfield   = move_stack[idx].ep_field;
    promoted  = move_stack[idx].promoted_fig;
    last_move = idx;

    // correct pawn variables depending on 'from' field contents
    switch(board[from]) {
        case W_PAWN:
            p_controlled[from+9].white--;
            p_controlled[from+11].white--;
            p_per_line[from%10].white--;
            if (promoted == EMPTY) {
                p_controlled[to+9].white++;
                p_controlled[to+11].white++;
                p_per_line[to%10].white++;
                break;
            }
            break;

        case B_PAWN:
            p_controlled[from-9].black--;
            p_controlled[from-11].black--;
            p_per_line[from%10].black--;
            if (promoted == EMPTY) {
                p_controlled[to-9].black++;
                p_controlled[to-11].black++;
                p_per_line[to%10].black++;
            }
            break;

        case B_ROOK:
            r_per_line[from%10].black--;
            r_per_line[to%10].black++;
            break;

        case W_ROOK:
            r_per_line[from%10].black--;
            r_per_line[to%10].white++;
            break;

        default:
            break;
    }

    if (epfield!=ILLEGAL && move_stack[idx].killed_fig==W_PAWN) {
        figure = -act_color;
        tmp_to = epfield;
    }
    else {
        figure = board[to];
        tmp_to = to;
    }

    // correct pawn variables depending on 'to' field contents
    switch(figure) {
        case W_PAWN:                    // killed white pawn
            p_controlled[tmp_to+9].white--;
            p_controlled[tmp_to+11].white--;
            p_per_line[tmp_to%10].white--;
            break;

        case B_PAWN:                     // killed black pawn
            p_controlled[tmp_to-9].black--;   
            p_controlled[tmp_to-11].black--;
            p_per_line[tmp_to%10].black--;
            break;

        case B_ROOK:                     // killed black rook
            r_per_line[tmp_to%10].black--;
            break;

        case W_ROOK:                     // killed white rook
            r_per_line[tmp_to%10].white--;
            break;

        default:
            break;
    }

    // one step deeper in the tree ...
    act_depth++;
    to_fields[act_depth] = to;
    ep_field[act_depth]  = ILLEGAL;

    // invert material balance to see it from the actual side
    material_bal[act_depth] = -material_bal[act_depth-1];
    material_sum[act_depth] = material_sum[act_depth-1];

    // figure is moving from the from-field to the to field
    move_indicators[from] = move_indicators[from]+1;
    move_indicators[to]   = move_indicators[to]+1;

    if (epfield != ILLEGAL) {
        if (board[epfield] == EMPTY) {    // pawn moves from 2nd to 4th row
            ep_field[act_depth] = epfield;
        }
        else {  // foreign pawn will be killed "enpassant"
            board[epfield] = EMPTY;
            material_bal[act_depth] = material_bal[act_depth] - VALUE_PAWN;
        }
    }

    else {
        // if a figure gets killed, change material balance
        if (move_stack[idx].killed_fig != EMPTY) {
            // a figure gets killed
            matchange = fig_values[move_stack[idx].killed_fig];
            material_bal[act_depth] = material_bal[act_depth] - matchange;
            // only figure (no pawns) gets counted in sum
            if (matchange != VALUE_PAWN) {
                material_sum[act_depth] = material_sum[act_depth] - matchange;
            }
        }
    }

    // execute move on board
    board[to]   = board[from];
    board[from] = EMPTY;

    // handle pawn promotion
    if (promoted != EMPTY) {
        board[to] = act_color*promoted;
        matchange = (fig_values[promoted]-VALUE_PAWN);
        material_bal[act_depth] = material_bal[act_depth] - matchange;
        // pawns will not be counted in sum
        material_sum[act_depth] = material_sum[act_depth] + matchange + VALUE_PAWN;
        // correct pawn variables for promotion
        if (promoted == W_ROOK) {
            if (act_color == WHITE) {    
                r_per_line[to%10].white++;
            }
            else {            
                r_per_line[to%10].white++;
            }
        }
    }
    else {
        // handle long rochade if necessary
        if (move_stack[idx].rochade_nr == ROCHADE_SHORT) {
            board[to+1] = EMPTY;
            board[to-1] = act_color*W_ROOK;
            rochade_indicators[act_color+1] = TRUE;
            // pawn variables correction for short rochade
            if (act_color == WHITE) {
                r_per_line[(to+1)%10].white--;
                r_per_line[(to-1)%10].white++;
            }
            else {
                r_per_line[(to+1)%10].black--;
                r_per_line[(to-1)%10].black++;
            }
        }
        // handle short rochade if necessary
        else if (move_stack[idx].rochade_nr == ROCHADE_LONG) {
            // pawn variables correction for long rochade
            if (act_color == WHITE) {
                r_per_line[(to-2)%10].white--;
                r_per_line[(to+1)%10].white++;
            }
            else {
                r_per_line[(to-2)%10].black--;
                r_per_line[(to+1)%10].black++;
            }
            board[to-2] = EMPTY;
            board[to+1] = act_color*W_ROOK;
            rochade_indicators[act_color+1] = TRUE;
        }
    }

    // update king positions if necessary
    if (board[to] == W_KING)    white_king_pos=to;
    else if (board[to]==B_KING) black_king_pos=to;

    // change color (other side is up to draw )
    act_color = -act_color;
}


/*===========================================================================*/
/* takes back a move                                                         */
/*===========================================================================*/
void UndoMove(int idx) {
    int from,to,epfield,promoted;
    int tmp_to,figure,rochade,fromfig;

    move_count--;
    from     = move_stack[idx].from;
    to       = move_stack[idx].to;
    epfield  = move_stack[idx].ep_field;
    promoted = move_stack[idx].promoted_fig;
    rochade  = move_stack[idx].rochade_nr;

    act_depth--;
    act_color   = -act_color;
    board[from] = board[to];
    board[to]   = EMPTY;

    if (epfield!=ILLEGAL && move_stack[idx].killed_fig==W_PAWN) {
        board[epfield] = -act_color; /* W_PAWN==WHITE, B_PAWN==BLACK */
    }

    // re-set killed figure
    else if (move_stack[idx].killed_fig != EMPTY) {
        board[to] = (-act_color)*move_stack[idx].killed_fig;
    }

    // decrement move indicators
    move_indicators[from] = move_indicators[from]-1;
    move_indicators[to]   = move_indicators[to]-1;

    // take back short rochade
    if (rochade == ROCHADE_SHORT) {
        board[to+1] = act_color*W_ROOK;
        board[to-1] = EMPTY;
        rochade_indicators[act_color+1]=FALSE;

        // correct pawn variables for short rochade
        if (act_color == WHITE) {
            r_per_line[(to-1)%10].white--;
            r_per_line[(to+1)%10].white++;
        }
        else {
            r_per_line[(to-1)%10].black--;
            r_per_line[(to+1)%10].black++;
        }    
    }
    // take back long rochade
    else if(rochade == ROCHADE_LONG) {
        board[to-2] = act_color*W_ROOK;
        board[to+1] = EMPTY;
        rochade_indicators[act_color+1] = FALSE;

        // correct pawn variables for long rochade
        if (act_color == WHITE) {
            r_per_line[(to-2)%10].white++;
            r_per_line[(to+1)%10].white--;
        }
        else {
            r_per_line[(to-2)%10].black++;
            r_per_line[(to+1)%10].black--;
        }
    }

    // take back pawn promotion
    if (promoted != EMPTY) board[from]=act_color;

    fromfig = board[from];

    // update king position if necessary
    if (fromfig==W_KING)      white_king_pos=from;
    else if (fromfig==B_KING) black_king_pos=from;


    //----------------------------------------------
    // handle the from statement for pawn structures
    //----------------------------------------------
    switch(fromfig) {
        case W_PAWN:
            p_controlled[from+9].white++;
            p_controlled[from+11].white++;
            p_per_line[from%10].white++;
            if (promoted == EMPTY) {
                p_controlled[to+9].white--;
                p_controlled[to+11].white--;
                p_per_line[to%10].white--;
            }
            break;

        case B_PAWN:
            p_controlled[from-9].black++;
            p_controlled[from-11].black++;
            p_per_line[from%10].black++;
            if (promoted == EMPTY) {
                p_controlled[to-9].black--;
                p_controlled[to-11].black--;
                p_per_line[to%10].black--;
            }
            break;

        case B_ROOK:
            r_per_line[from%10].black++;
            if (promoted != W_ROOK) {
                r_per_line[to%10].black--;
            }
            break;

        case W_ROOK:
            r_per_line[from%10].white++;
            if (promoted != W_ROOK) {
                r_per_line[to%10].white--;
            }
            break;

        default:
            break;
    }


    if (epfield!=ILLEGAL && move_stack[idx].killed_fig==W_PAWN) {
        figure = -act_color;
        tmp_to = epfield;
    }
    else {
        figure = board[to];
        tmp_to = to;
    }
                
    
    //------------------------------------------------
    // handle the killed statement for pawn structures
    //------------------------------------------------
    switch(figure) {
        case W_PAWN:                    // killed white pawn
            p_controlled[tmp_to+9].white++;
            p_controlled[tmp_to+11].white++;
            p_per_line[tmp_to%10].white++;
            break;

        case B_PAWN:                     // killed black pawn
            p_controlled[tmp_to-9].black++;   
            p_controlled[tmp_to-11].black++;
            p_per_line[tmp_to%10].black++;
            break;

        case B_ROOK:                      // killed black rook
            r_per_line[tmp_to%10].black++;
            break;

        case W_ROOK:                     // killed white rook
            r_per_line[tmp_to%10].white++;
            break;

        default:
            break;
    }

}




/*===========================================================================*/
/* evaluation positional value of a black pawn                               */
/*                                                                           */
/* This is one of the most called functions at the beginning ...             */
/*===========================================================================*/
inline int BlackPawnEvaluation(int field,int row, int column,int developed) {
    register int value;
    register int pctrl;
    register int pctrl10;
    register int rowX2;
    register int j;

#if DO_STATISTICS
    bpevalcount++;
#endif

    // invert row. higher row is better (equal to white)
    row = (ROW_8+ROW_1)-row;
    rowX2 = row >> 1;

    // opening or middle game ?
    if (material_sum[act_depth]>VALUE_ENDGAME) {
        value = bp_fieldvalues[field];

        // development not finished. don't force border pawns
        if (developed<4  && 
            (column>=F_COLUMN || column<=B_COLUMN) &&
            row > ROW_3)
        {
            value -= 15;
        }
    }
    else {
        // if we are not in the endgame all lines are treated equally
        value = rowX2;
    }

    // malus for "isolanis"
    if (p_per_line[column-1].black==0 &&
        p_per_line[column+1].black==0) 
    {
        value -= 12;
        // additional malus for isolated double pawns on a line
        if (p_per_line[column].black>1) value -= 12;
    }

    // malus for double pawns on a line
    if (p_per_line[column].black>1) value -= 15;

    pctrl   = p_controlled[field].black;
    pctrl10 = p_controlled[field+10].black;

    // bonus for "duos" (for example: e5/d5)
    if (pctrl>0 || pctrl10>0) {
        value += row ;
    }

    if (p_per_line[column].white==0) {
        if (pctrl==0 &&
            p_controlled[field-10].white > pctrl10) 
        {
            value -= 10;
            if (r_per_line[column].white>0) value -= 8; 
        }
        else {
            for (j=field;j>H2;j-=10) {
                if (p_controlled[j].white>0) {
                    goto finished;
                }
            }
            // free pawn found. will be weighted higher in endgame
            if (material_sum[act_depth]<VALUE_ENDGAME) {
                value += row>>4; // bonus for wide processed pawn
                if (r_per_line[column].black>0) { // rook supports pawn on same line
                    value += rowX2;
                }
                if (r_per_line[column].white>0) { // foreign rook on same line
                    value -= rowX2;   
                }
                if (material_sum[act_depth]==0) { // only pawns left
                    value += row>>3; // extra bonus for position
                }
                if (pctrl>0 || pctrl10>0) { // supported pawn bonus
                    //value += row>>2;
                    value += rowX2;
                    value += rowX2;
                }
                // malus for blocked pawn
                if (pctrl10 == 0 && board[field-10]<0) {
                    //value -= row>>2;
                    value -= rowX2;
                    value -= rowX2;
                }
            }
            else { // free pawn in middle game
                value += row>>3;
                if (pctrl>0 || pctrl10>0) { // supported pawn bonus
                    value += row>>2;
                }
            }
        }
    }

finished:
    return value;
}


/*===========================================================================*/
/* evaluation positional value of a white pawn                               */
/*                                                                           */
/* This is one of the most called functions at the beginning ...             */
/*===========================================================================*/
inline int WhitePawnEvaluation(int field,int row, int column,int developed) {
    register int value;
    register int pctrl;
    register int pctrl10;
    register int rowX2;
    register int j;

#if DO_STATISTICS
    wpevalcount++;
#endif

    rowX2 = row >> 1;

    if (material_sum[act_depth]>VALUE_ENDGAME) {
        value = wp_fieldvalues[field];
        if (developed<4 &&
            (column>=F_COLUMN || column<=B_COLUMN) &&
            row>ROW_3) 
        {
            value -= 15;
        }
    }
    else {
        value = row>>2;
    }

    if (p_per_line[column-1].white==0 && 
        p_per_line[column+1].white==0) 
    {
        value -= 12;  // Isolani 
        if (p_per_line[column].white>1) { // isolated double pawn
            value -= 12;
        }
    }

    if (p_per_line[column].white>1) { // double pawn
        value -= 15;
    }
    
    pctrl   = p_controlled[field].white;
    pctrl10 = p_controlled[field+10].white;

    // duo or supporting pawn bonus
    if (pctrl>0 || pctrl10>0) {
        value += row;
    }
    
    if (p_per_line[column].black==0) { // halfopen line
        // not much developed pawn on halfopen line
        if (pctrl==0 &&
            p_controlled[field+10].black>pctrl10) 
        {
            value -= 10;
            if (r_per_line[column].black>0) {
                value -= 8; // rook "attacks" pawn
            }
        }
        else {
            for (j=field;j<A7;j+=10) { // check for free pawn
                if (p_controlled[j].black>0) {
                    goto finished2;
                }
            }
            // free pawn found 
            if (material_sum[act_depth]<VALUE_ENDGAME) {
                value += row>>4;

                if (r_per_line[column].white>0) { // supporting rook
                    value += rowX2;
                }
                if (r_per_line[column].black>0) { // foreign rook
                    value -= rowX2;
                }

                if (material_sum[act_depth]==0) { // only pawns left
                    value += row>>3;
                }

                if (pctrl>0 || pctrl10>0) { // supporting free pawn
                    value += rowX2;
                    value += rowX2;
                    //value += row>>2;
                }

                // blocked by black figure
                if (pctrl10==0 && board[field+10]<0) {
                    //value -= row>>2;
                    value -= rowX2;
                    value -= rowX2;
                }
            }
            else { // free pawn in middle game
                value += row>>3;
                if (pctrl>0 || pctrl10>0) { // supporting free pawn
                    value += rowX2;
                }
            }
        }
    }

finished2:
    return value;
}


/*===========================================================================*/
/* board evaluation from the view of "side"                                  */
/*                                                                           */
/* if material value is too much out of alpha/beta window, just the material */
/* is evaluated.                                                             */
/*                                                                           */
/* if "side" will generate a MATT function returns VALUE_MATT-depth          */
/*===========================================================================*/
int Evaluate(int alpha,int beta,int side) {
    int value;           // material balance
    int posvalue;        // positional evaluation
    int i,j,k,field;
    int white_bishops;   // counts white bishops
    int black_bishops;   // counts black bishops
    int pawn_count;      // pawns on the board (white+black)
    int matsum;          // material sum short cut
    int white_rooks_on7; // white rooks on 7/6 row counter
    int black_rooks_on2; // black rooks on 1/2 row counter
    int white_developed; // development state of white (0..5)
    int black_developed; // development state of black (0..5)

    int my_king_pos,foreign_king_pos;

    //------------------------------------------------------
    // check first if opponent is MATT or 'side' is in chess
    //------------------------------------------------------
    if (side==WHITE) {
        my_king_pos      = white_king_pos;
        foreign_king_pos = black_king_pos;
    }
    else {
        my_king_pos      = black_king_pos;
        foreign_king_pos = white_king_pos;
    }

    if (AttacksField(foreign_king_pos,side)==TRUE) return VALUE_MATT-act_depth;
    chess_indicator=AttacksField(my_king_pos,side*(-1));

    //--------------------------------------------------------------------
    // positional factors won't count if material balance is too unequally
    // EXCEPTION: late endgame
    //--------------------------------------------------------------------
    value   = material_bal[act_depth];
    matsum  = material_sum[act_depth];

    if (matsum > (2*VALUE_BISHOP)) {
        if ((value<alpha-MAX_POS) || (value>beta+MAX_POS)) return value;
    }

    //InitPawnVars();

#if CHECK_PAWN_STRUCT
    CheckPawnVars();
#endif

    posvalue        = 0;
    black_bishops   = 0;
    white_bishops   = 0;
    pawn_count      = 0;
    white_rooks_on7 = 0;
    black_rooks_on2 = 0;

    //--------------------------------------------------------------
    // evaluate development state of white and black by checking
    // for rochade and light figure movement
    //
    // POSSIBLE IMPROVEMENT: make develop vars global and check only
    //                       if value < 5.
    //                       !! BUT BE AWARE OF TEMPORARY MOVES !!
    //--------------------------------------------------------------
    white_developed=rochade_indicators[WHITE+1];
    if (move_indicators[B1]>0) white_developed++;  // knight B1 moved?
    if (move_indicators[C1]>0) white_developed++;  // bishop C1 moved?
    if (move_indicators[F1]>0) white_developed++;  // bishop F1 moved?
    if (move_indicators[G1]>0) white_developed++;  // knight G1 moved?

    black_developed=rochade_indicators[BLACK+1];
    if (move_indicators[B8]>0) black_developed++;  // knight B8 moved?
    if (move_indicators[C8]>0) black_developed++;  // bishop C8 moved?
    if (move_indicators[F8]>0) black_developed++;  // bishop F8 moved?
    if (move_indicators[G8]>0) black_developed++;  // knight G8 moved?

    //------------------------------------------------------
    // loop through complete board and evaluate each figure.
    // evaluation is done from WHITE's sight
    //
    // posvalue > 0 ... better for white
    // posvalue < 0 ... better for black
    //------------------------------------------------------
    for (i=ROW_1;i<=ROW_8;i++) {
        field = i*10 + A_COLUMN;
        for (j=A_COLUMN;j<=H_COLUMN;j++,field++) {
            switch(board[field]) {
                case B_KING:
                    //-----------------------------
                    // endgame evaluation
                    //-----------------------------
                    if (matsum<VALUE_ENDGAME) {
                        posvalue -= center_table[field];
                    }
                    else {
                        // malus for rochade right lost
                        if ((rochade_indicators[BLACK+1]==FALSE) &&
                            ((move_indicators[E8]>0) ||
                            ((move_indicators[H8]>0) &&
                             (move_indicators[A8]>0))))
                        {
                            posvalue += 35;
                        }

                        // king shouldn't be in the center
                        posvalue += (4*center_table[field]);

                        for (k=-1;k<=1;k++) {
                            // bonus for pawns directly in front of king
                            if (board[field-10+k] == B_PAWN) posvalue -= 15;
                            // bonus for pawns 2 rows in front of king
                            if (board[field-20+k] == B_PAWN) posvalue -= 6;
                            // malus for half-open line with a foreign rook
                            if (p_per_line[j+k].white==0 && r_per_line[j+k].white>0) {
                                posvalue=posvalue+12;
                            }
                        }
                    }
                    break;

                case B_QUEEN:
                    // malus for queen excursations at the beginning
                    if (black_developed<4) {
                        if (field<A8) posvalue += 15;
                    }
                    // bonus if queen is near foreign king
                    else {
                        posvalue += 2*(abs((white_king_pos/10)-(field/10)) +
                                       abs((white_king_pos%10)-(field%10)));
                    }
                    break;

                case B_KNIGHT:
                    // should be near the center
                    posvalue -= (center_table[field]/2);
                    break;

                case B_BISHOP:
                    // shouldn't block d7/e7 pawns
                    if ((field==D6 || field==E6) &&
                        (board[field+10]==B_PAWN))
                    {
                        posvalue += 20;
                    }
                    black_bishops++;
                    break;

                case B_ROOK:
                    // bonus if it is on row 1 or 2
                    if (field<=H2) black_rooks_on2++;

                    // bonus if near center
                    if (j>=C_COLUMN && j<=E_COLUMN) posvalue -= 4;

                    // bonus if on halfopen or open line
                    if (p_per_line[j].white==0) {
                        posvalue-=8;
                        if(p_per_line[j].black==0) posvalue-=5;
                    }
                    break;

                case B_PAWN:
                    // will be done in own routine
                    posvalue -= BlackPawnEvaluation(field,i,j,black_developed);
                    pawn_count++;
                    break;

                case EMPTY:
                    break;

                case W_PAWN:
                    // will be done in own routine
                    posvalue += WhitePawnEvaluation(field,i,j,white_developed);
                    pawn_count++;
                    break;

                case W_ROOK:
                    // bonus if on row 7 or 8
                    if (field>=A7) white_rooks_on7++;

                    // bonus if near center
                    if (j>=C_COLUMN && j<=E_COLUMN) posvalue += 4;

                    // bonus if on halfopen or open line
                    if (p_per_line[j].black==0) {
                        posvalue+=8;
                        if (p_per_line[j].white==0) posvalue += 5;
                    }
                    break;

                case W_BISHOP:
                    // malus if blocking d3 or e3 pawn
                    if ((field==D3 || field==E3) &&
                        (board[field-10]==W_PAWN))
                    {
                        posvalue-=20;
                    }
                    white_bishops++;
                    break;

                case W_KNIGHT:
                    posvalue+=(center_table[field]/2);
                    break;

                case W_QUEEN:
                    // malus for queen excursations at the beginning
                    if (white_developed<4) {
                        if (field>H1) posvalue-=15;
                    }
                    // queen should be near foreign king
                    else {
                        posvalue -= 2*(abs((black_king_pos/10)-(field/10)) +
                                       abs((black_king_pos%10)-(field%10)));
                    }
                    break;

                case W_KING:
                    // in endgame king should be in the center
                    if (matsum<VALUE_ENDGAME) {
                        posvalue+=center_table[field];

                        // handle near opposition of kings
                        if (abs(field-black_king_pos)==20 || abs(field-black_king_pos)==2) {
                            k=10;
                            if (matsum==0) k=30;

                            if (act_color==WHITE) posvalue -= k;
                            else                  posvalue += k;
                        }
                    }
                    else {
                        // malus for lost rochade right
                        if ((rochade_indicators[WHITE+1]==FALSE)&&
                            ((move_indicators[E1]>0)||
                            ((move_indicators[H1]>0)&&
                            (move_indicators[A1]>0))))
                        {
                            posvalue -= 35;
                        }
                        // king shouldn't be in the center
                        posvalue -= (4*center_table[field]);
                        for (k=-1;k<=1;k++) {
                            // bonus for own pawn directly in front of king
                            if (board[field+10+k]==W_PAWN) posvalue+=15;
                            // bonus for own pawn 2 rows in front of king
                            if (board[field+20+k]==W_PAWN)  posvalue+=6;
                            // malus for halfopened line with foreign rook
                            if (p_per_line[j+k].black==0 && r_per_line[j+k].black>0) {
                                posvalue -= 12;
                            }
                        }
                    }
                    break;
            }
        }
    }

    //------------------------------------------------------------
    // try to recognize insufficient material on board
    // will recognize: (K,K),(KB,K) (KN,K) (KNN,K) (KB,KB) (KB,KN)
    //
    // BUT NOT: (KNN,KB)  --- will this always be "PATT" ???
    //------------------------------------------------------------
    if (pawn_count==0) {
        if ((matsum<=VALUE_BISHOP) || (matsum==2*VALUE_KNIGHT) ||
            ((abs(material_bal[act_depth])<100)&&
             (matsum<=(2*VALUE_BISHOP))))
        {
            return 0;
        }
    }

    // bonus for white if both bishops are living
    if (white_bishops>=2) posvalue+=15;

    // bonus for black if both bishops are living
    if (black_bishops>=2) posvalue-=15;

    // extra bonus for white rooks
    if (white_rooks_on7>0 && black_king_pos>=A7) {
        posvalue+=10;
        if (white_rooks_on7>1) posvalue+=25;
    }

    // extra bonus for black rooks
    if (black_rooks_on2>0 && white_king_pos<=H2) {
        posvalue-=10;
        if (black_rooks_on2>1) posvalue-=25;
    }

    // if side is BLACK --> invert position evaluation value
    if (side==BLACK) posvalue = -posvalue;

    // take mobility of knights and rooks into account
    // (will be calculated by move generator)
    if (act_depth>=1) {
        posvalue -= (mobility[act_depth]-mobility[act_depth-1])/16;
    }

    return value+posvalue;
}


/*===========================================================================*/
/* stores a move on the move stack                                           */
/*===========================================================================*/
void PutMoveOnStack(int from,int to) {
    //------------------------------------------------------
    // increment mobility for bishops and rooks of own color
    // mobility near center will be weighted higher than
    // near the border
    //------------------------------------------------------
    register int     figure;
    register move_t* tmpmove;

    figure = board[from];

    if (act_color==WHITE) {
        if (figure == W_BISHOP || figure == W_ROOK) {
            mobility[act_depth] += center_table[to];
        }
    }
    else {
        if (figure == B_BISHOP || figure == B_ROOK) {
            mobility[act_depth] += center_table[to];
        }
    }

    tmpmove = &move_stack[act_index];

    //----------------------------------------
    // weighting of move for move sorting
    // bonus for main variant and killer moves
    //----------------------------------------
    if (main_var[0][act_depth].from==from &&
        main_var[0][act_depth].to==to)
    {
        tmpmove->value=HVARBONUS;
    }
    else if (killer_tab[act_depth].killer1.from==from &&
             killer_tab[act_depth].killer1.to==to)
    {
        tmpmove->value=KILLER1BONUS;
    }
    else if (killer_tab[act_depth].killer2.from==from &&
             killer_tab[act_depth].killer2.to==to)
    {
        tmpmove->value=KILLER2BONUS;
    }
    else {
        tmpmove->value=EMPTY;
    }

    tmpmove->from         = from;
    tmpmove->to           = to;
    tmpmove->killed_fig   = EMPTY;
    tmpmove->promoted_fig = EMPTY;
    tmpmove->rochade_nr   = ROCHADE_NONE;
    tmpmove->ep_field     = ILLEGAL;

    // increment act_index and check for overflow
    if (act_index<STACK_SIZE) {
        act_index++;
    }
    else {
        LeavePrg("ERROR: move stack overflow\n");
    }
}


/*===========================================================================*/
/* stores a killing move on the move stack                                   */
/*===========================================================================*/
void PutKillMoveOnStack(int from,int to) {
    register int     figvalue;
    register move_t* tmpmove;

    figvalue = board[to];

    // cannot kill a king
    if (figvalue == W_KING || figvalue == B_KING) return;

    figvalue=fig_values[abs(figvalue)];

    tmpmove = &move_stack[act_index];

    tmpmove->from       = from;
    tmpmove->to         = to;
    tmpmove->killed_fig = abs(board[to]);

    // bonus for low value figures killing higher level figures
    tmpmove->value = figvalue-(fig_values[abs(board[from])]/8);

    // extra bonus for killing last moved figure
    if (act_depth>0 && to==to_fields[act_depth-1]) {
        tmpmove->value += 300;
    }

    // bonus for main variant and killer moves
    if (main_var[0][act_depth].from==from &&
        main_var[0][act_depth].to==to)
    {
        tmpmove->value += HVARBONUS;
    }
    else if (killer_tab[act_depth].killer1.from==from &&
             killer_tab[act_depth].killer1.to==to)
    {
        tmpmove->value += KILLER1BONUS;
    }
    else if (killer_tab[act_depth].killer2.from==from &&
             killer_tab[act_depth].killer2.to==to)
    {
        tmpmove->value += KILLER2BONUS;
    }
    tmpmove->promoted_fig = EMPTY;
    tmpmove->rochade_nr   = ROCHADE_NONE;
    tmpmove->ep_field     = ILLEGAL;

    // increment act_index and check for overflow
    if(act_index<STACK_SIZE) {
        act_index++;
    }
    else {
        LeavePrg("ERROR: move stack overflow\n");
    }
}


/*===========================================================================*/
/* stores enpassant move on the move_stack                                   */
/*===========================================================================*/
void PutEpOnStack(int from,int to,int epfield) {
    register int     figure;
    register move_t* tmpmove;

    figure = board[to];

    // kings cannot be killed
    if (figure == W_KING || figure == B_KING) return;

    tmpmove = &move_stack[act_index];

    tmpmove->from         = from;
    tmpmove->to           = to;
    tmpmove->killed_fig   = W_PAWN;
    tmpmove->promoted_fig = EMPTY;
    tmpmove->rochade_nr   = ROCHADE_NONE;
    tmpmove->ep_field     = epfield;
    tmpmove->value        = VALUE_PAWN;

    // increment act_index and check for move stack overflow (shouldn't happen)
    if (act_index<STACK_SIZE) {
        act_index++;
    }
    else {
        LeavePrg("ERROR: move stack overflow\n");
    }
}


/*===========================================================================*/
/* stores all possible promotions on the move stack                          */
/*===========================================================================*/
void PutPromotionOnStack(int from,int to) {
    int i;
    if (board[to]==EMPTY) {
        for (i=W_QUEEN;i>W_PAWN;i--) { // ordering: Q/N/B/R
            PutMoveOnStack(from,to);
            move_stack[act_index-1].promoted_fig = i;
        }
    }
    else {
        for (i=W_QUEEN;i>W_PAWN;i--) {
            PutKillMoveOnStack(from,to);
            move_stack[act_index-1].promoted_fig = i;
        }
    }
}


/*===========================================================================*/
/* checks if given side attacks given field                                  */
/*                                                                           */
/* returns TRUE if field is under attack, otherwise FALSE                    */
/*                                                                           */
/* method: simulates a king of metafigure which may attack in all directions */
/*===========================================================================*/
int AttacksField(int field, int side) {
    int i,direction,to,figure,slide;

#if DO_STATISTICS
    attackcount++;
#endif

    //-------------------
    // !!!! SPEED UP !!!!
    //-------------------
    if (side == WHITE) {
        // test for pawn attack ...
        if (board[field-9]==W_PAWN || board[field-11]==W_PAWN) return TRUE;
        // test for knight attack ...
        for (i=8;i<16;i++) {
            figure = board[field+offset_vals[i]];        
            if (figure==W_KNIGHT) return TRUE;
        }
    }
    else {
        // test for pawn attack ...
        if (board[field+9]==B_PAWN || board[field+11]==B_PAWN) return TRUE;
        // test for knight attack ...
        for (i=8;i<16;i++) {
            figure = board[field+offset_vals[i]];        
            if (figure==B_KNIGHT) return TRUE;
        }
    }
    
    //------------------------------------------------------
    // test for sliding figures and king in all 8 directions
    //------------------------------------------------------
    for (i=0;i<8;i++) {
        to        = field;
        direction = offset_vals[i];
        slide     = 0;
slideforward:
        slide++;
        to+=direction;

        figure = board[to];
        

        if (figure==EMPTY)   goto slideforward; // slide again        
        if (figure==OUTSIDE) continue;          // oops, outside !

        if (side==WHITE) {
            if (figure>0) {
                if (figure==W_KING) {
                    if (slide<=1) return TRUE;
                }
                else {
                    if (figure_offsets[figure].start<=i &&
                        figure_offsets[figure].end>=i)
                    {
                        return TRUE;
                    }
                }
           }
        }
        else {  // for black
            if (figure<0) {
                if (figure==B_KING) {
                    if (slide<=1) return TRUE;
                }
                else {
                    if (figure_offsets[-figure].start<=i &&
                        figure_offsets[-figure].end>=i)
                    {
                        return TRUE;
                    }
                }
            }
        }
    }

    // everything checked. nothing found.
    return FALSE;
}


/*===========================================================================*/
/* generates moves and stores them on move_stack                             */
/*                                                                           */
/* if allmoves > 0 --> all pseudo legal moves are generated                  */
/* else            --> just generate promotion, enpassant, rochade and       */
/*                     killing moves                                         */
/*===========================================================================*/
void GenerateMoves(int allmoves) {
    int from,to,figure;
    int i,longstep,direction;
    int epfield;

    act_index=stack_ranges[act_depth]; // set act_index to begin of act. range
    mobility[act_depth] = 0;

    // loop through complete board
    for (from=A1;from<=H8;from++) {
        figure = board[from];
        if (figure==EMPTY || figure==OUTSIDE) continue;

        // figure have to be of right color
        if((act_color==WHITE && figure<0) ||
           (act_color==BLACK && figure>0))
        {
            continue;
        }

        figure=abs(figure);   // kind of figure only matters with pawns

        if (figure==W_PAWN) {
            //----------------------------------------
            // treatment of white pawn ...
            //----------------------------------------
            if (act_color==WHITE) {
                if (board[from+10]==EMPTY) {
                    if (from>=A7) {
                        PutPromotionOnStack(from,from+10);
                    }
                    else if (allmoves>0) {
                        PutMoveOnStack(from,from+10);
                        // double step possible ?
                        if (from<=H2 && board[from+20]==EMPTY) {
                            PutMoveOnStack(from,from+20);
                            // put move on stack have already incremented idx
                            move_stack[act_index-1].ep_field=from+10;
                        }
                    }
                }
                if (board[from+11]<0) { // pawn may kill black figure
                    if (from>=A7) {
                        PutPromotionOnStack(from,from+11);
                    }
                    else {
                        PutKillMoveOnStack(from,from+11);
                    }
                }

                if (board[from+9]<0) {  // other direction, too
                    if (from>=A7) {
                        PutPromotionOnStack(from,from+9);
                    }
                    else {
                        PutKillMoveOnStack(from,from+9);
                    }
                }
            }

            //----------------------------------------
            // treatment of black pawn ...
            //----------------------------------------
            else if (act_color==BLACK) {
                if (board[from-10]==EMPTY) {
                    if (from<=H2) {
                        PutPromotionOnStack(from,from-10);
                    }
                    else if(allmoves>0) {
                        PutMoveOnStack(from,from-10);
                        // double step possible?
                        if ((from>=A7)&&(board[from-20]==EMPTY)) {
                            PutMoveOnStack(from,from-20);
                            // putting move on stack already incremented idx
                            move_stack[act_index-1].ep_field=from-10;
                        }
                    }
                }
                // have to check for border just for black pawns
                if (board[from-11]>0 && board[from-11]!=OUTSIDE) {
                    if (from<=H2) {
                        PutPromotionOnStack(from,from-11);
                    }
                    else {
                        PutKillMoveOnStack(from,from-11);
                    }
                }
                if (board[from-9]>0 && board[from-9]!=OUTSIDE) {
                    if (from<=H2) {
                        PutPromotionOnStack(from,from-9);
                    }
                    else {
                        PutKillMoveOnStack(from,from-9);
                    }
                }
            }
            continue;   // continue with next field
        }

        //------------------------------------------------
        // moves of all other figures are calculated using
        // the move offset table
        //------------------------------------------------
        longstep=figure_offsets[figure].longmove;
        for (i=figure_offsets[figure].start;i<=figure_offsets[figure].end;i++) {
            direction = offset_vals[i];
            to        = from;
slideforward:
            to += direction;
            if (board[to]==EMPTY) {
                if (allmoves>0) {
                    PutMoveOnStack(from,to);
                }
                if (longstep) { // bishop/rook/queen
                    goto slideforward;
                }
                else {  // knight/king
                    continue;
                }
            }
            if (board[to]==OUTSIDE) {
                continue;
            }

            // there is a figure. check for right color ...
            if ((act_color==WHITE && board[to]<0) ||
                (act_color==BLACK && board[to]>0))
            {
                PutKillMoveOnStack(from,to);
            }
            continue;
        }
    }

    //-------------------------------
    // enpassant
    //-------------------------------
    if (ep_field[act_depth] != ILLEGAL) {
        epfield=ep_field[act_depth];
        if (act_color==WHITE) {
            if (board[epfield-9]==W_PAWN) {
                PutEpOnStack(epfield-9,epfield,epfield-10);
            }
            if (board[epfield-11]==W_PAWN) {
                PutEpOnStack(epfield-11,epfield,epfield-10);
            }
        }
        else {
            if (board[epfield+9]==B_PAWN) {
                PutEpOnStack(epfield+9,epfield,epfield+10);
            }
            if (board[epfield+11]==B_PAWN) {
                PutEpOnStack(epfield+11,epfield,epfield+10);
            }
        }
    }

    //-------------------------------
    // rochade
    //-------------------------------
    if (act_color==WHITE) {
        if (white_king_pos==E1 && move_indicators[E1]==0) {
            if (board[H1]==W_ROOK             &&
                move_indicators[H1]==0        &&
                board[F1]==EMPTY              &&
                board[G1]==EMPTY              &&
                AttacksField(E1,BLACK)==FALSE &&
                AttacksField(F1,BLACK)==FALSE &&
                AttacksField(G1,BLACK)==FALSE)
            {
                PutMoveOnStack(E1,G1); // put king move on stack
                move_stack[act_index-1].rochade_nr=ROCHADE_SHORT;
            }
            if (board[A1]==W_ROOK             &&
                move_indicators[A1]==0        &&
                board[D1]==EMPTY              &&
                board[C1]==EMPTY              &&
                board[B1]==EMPTY              &&
                AttacksField(E1,BLACK)==FALSE &&
                AttacksField(D1,BLACK)==FALSE &&
                AttacksField(C1,BLACK)==FALSE)
            {
                PutMoveOnStack(E1,C1); // put king move on stack
                move_stack[act_index-1].rochade_nr=ROCHADE_LONG;
            }
        }
    }
    else {
        if (black_king_pos==E8 && move_indicators[E8]==0) {
            if (board[H8]==B_ROOK             && 
                move_indicators[H8]==0        &&
                board[F8]==EMPTY              && 
                board[G8]==EMPTY              &&
                AttacksField(E8,WHITE)==FALSE &&
                AttacksField(F8,WHITE)==FALSE &&
                AttacksField(G8,WHITE)==FALSE)
            {
                PutMoveOnStack(E8,G8); // put king move on stack
                move_stack[act_index-1].rochade_nr=ROCHADE_SHORT;
            }
            if (board[A8]==B_ROOK             &&
                move_indicators[A8]==0        &&
                board[D8]==EMPTY              &&
                board[C8]==EMPTY              &&
                board[B8]==EMPTY              &&
                AttacksField(E8,WHITE)==FALSE &&
                AttacksField(D8,WHITE)==FALSE &&
                AttacksField(C8,WHITE)==FALSE)
            {
                PutMoveOnStack(E8,C8); // put king move on stack
                move_stack[act_index-1].rochade_nr=ROCHADE_LONG;
            }
        }
    }
    stack_ranges[act_depth+1]=act_index; // store end of range
}


#if CHECK_PAWN_STRUCT

bicolor_t p_controlled2[BOARD_SIZE] = {};
bicolor_t p_per_line2[H_COLUMN+2]   = {};
bicolor_t r_per_line2[H_COLUMN+2]   = {};

void CheckPawnVars(void) {
    int i;
    char figure;

    // reset all counters to zero
    memset(&p_controlled2[A1],0,(H8-A1+1)*sizeof(bicolor_t));
    memset(p_per_line2,0,(H_COLUMN+2)*sizeof(bicolor_t));
    memset(r_per_line2,0,(H_COLUMN+2)*sizeof(bicolor_t));

    // evaluate new values
    for (i=A1;i<=H8;i++) {
        figure = board[i];

        switch(figure) {
            case EMPTY:   // fall through !!!
            case OUTSIDE: continue;
            case W_PAWN:
                p_controlled2[i+9].white++;
                p_controlled2[i+11].white++;
                p_per_line2[i%10].white++;
                break;

            case B_PAWN:
                p_controlled2[i-9].black++;
                p_controlled2[i-11].black++;
                p_per_line2[i%10].black++;
                break;

            case B_ROOK:
                r_per_line2[i%10].black++;
                break;

            case W_ROOK:
                r_per_line2[i%10].white++;
                break;

            default:
                break;
        }
    }

    for (i=A1;i<=H8;i++) {
        if (i<H_COLUMN+2) {
            if (p_per_line2[i].white != p_per_line[i].white) {printf("p_per[W] %d %d %d %s %d\n",i,p_per_line2[i].white,p_per_line[i].white,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
            if (p_per_line2[i].black != p_per_line[i].black) {printf("p_per[B] %d %d %d %s %d\n",i,p_per_line2[i].black,p_per_line[i].black,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
            if (r_per_line2[i].white != r_per_line[i].white) {printf("r_per[W] %d %d %d %s %d\n",i,r_per_line2[i].white,r_per_line[i].white,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
            if (r_per_line2[i].black != r_per_line[i].black) {printf("r_per[B] %d %d %d %s %d\n",i,r_per_line2[i].black,r_per_line[i].black,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
        }
        if (p_controlled2[i].white != p_controlled[i].white) {printf("p_cont[W] %d %d %d %s %d\n",i,p_controlled2[i].white,p_controlled[i].white,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
        if (p_controlled2[i].black != p_controlled[i].black) {printf("p_cont[B] %d %d %d %s %d\n",i,p_controlled2[i].black,p_controlled[i].black,Move2Str(&move_stack[last_move]),move_stack[last_move].ep_field);ngetchx();}
    }
} 

#endif

/*===========================================================================*/
/* fills up:                                                                 */
/*                                                                           */
/* (1) p_controlled ...  fields controlled by pawns                          */
/* (2) p_per_line   ...  number of pawns on a line                           */
/* (3) r_per_line   ...  number of rooks on a line                           */
/*                                                                           */
/* (will be called by Evaluate() to initialize)                              */
/*===========================================================================*/
void InitPawnVars(void) {
    int i;
    char figure;

    // reset all counters to zero
    memset(&p_controlled[A1],0,(H8-A1+1)*sizeof(bicolor_t));
    memset(p_per_line,0,(H_COLUMN+2)*sizeof(bicolor_t));
    memset(r_per_line,0,(H_COLUMN+2)*sizeof(bicolor_t));

    // evaluate new values
    for (i=A1;i<=H8;i++) {
        figure = board[i];

        switch(figure) {
            case EMPTY:   // fall through !!!
            case OUTSIDE: continue;
            case W_PAWN:
                p_controlled[i+9].white++;
                p_controlled[i+11].white++;
                p_per_line[i%10].white++;
                break;

            case B_PAWN:
                p_controlled[i-9].black++;
                p_controlled[i-11].black++;
                p_per_line[i%10].black++;
                break;

            case B_ROOK:
                r_per_line[i%10].black++;
                break;

            case W_ROOK:
                r_per_line[i%10].white++;
                break;

            default:
                break;
        }
    }
}


/*===========================================================================*/
/* calculates computer move. searching will be done iterative until          */
/* std_depth is reached.                                                     */
/*===========================================================================*/
move_t* ComputerMove(void) {
    int    distance,alpha,beta,i,value,bestvalue,j;
    int    chess_tmp;
    move_t tmp;
    int    moves_in_line;
    char   s[100];

    if (already_matt == TRUE) return 0;

    InitTree();

#if DO_STATISTICS
    nodescount+=nr_nodes;
#endif
    updatecounter=0;

    // evaluate initial position. if alreay matt -> stop

    value = Evaluate(-VALUE_MATT,VALUE_MATT,act_color);
    if (value==VALUE_MATT) {
        already_matt = TRUE;
        return 0;
    }

    // store chess_indicator for patt state detection at the end
    chess_tmp=chess_indicator;
    nr_nodes=0;

    // generate all pseudo legal moves
    GenerateMoves(1);


    // initialize pawn structures
    InitPawnVars();

    //------------------------------------------------------------------------
    // possible later speedup: remove all illegal moves from the move stack ??
    //------------------------------------------------------------------------

    // iterative stepping down until std_depth is reached
    for (distance=1;distance<=std_depth;distance++) {
        // at act_depth 1 the alpha/beta window is completely open
        if(distance==1) { 
            alpha = -VALUE_MATT; 
            beta  =  VALUE_MATT; 
        }
        // otherwise modify window a little bit
        else {              
            beta  = alpha+100;  
            alpha = alpha-100; 
        }
        // search will be extended for chess and killing moves
        max_extend    = distance+3;
        moves_in_line = 0;

        //evaluate value of each move
        for (i=0;i<stack_ranges[1];i++) {
            // stop calculation if a key is pressed
            if (kbhit()) {
                int answer;
                ngetchx();
                puts("\nAbort (Y/N)?");
                do {
                    answer = toupper(ngetchx());
                    if (answer == 'Y') {
                        puts("Y\n");
                        return 0;
                    }
                }
                while (answer != 'N'); 
                puts("N\n");
            }
            moves_in_line++;

            // initialize main variant
            main_var[act_depth][act_depth].from=0; 

            // execute move, evaluate move and take it back
            ExecMove(i);
            value = -AlphaBetaSearch(-beta,-alpha,distance-1);
            UndoMove(i);
            
            // we need an exact result for the first one
            if (i==0) { 
                if (value<alpha) {
                    // search for best move falls bellow lower window border
                    // --> new search with extended border
                    alpha         = -VALUE_MATT;
                    beta          = value;
                    moves_in_line = 0;
                    ExecMove(i);
                    value = -AlphaBetaSearch(-beta,-alpha,distance-1);
                    UndoMove(i);
                }
                else if(value>=beta) { // higher than upper window border
                    alpha         = value;
                    beta          = VALUE_MATT;
                    moves_in_line = 0;
                    ExecMove(i);
                    value = -AlphaBetaSearch(-beta,-alpha,distance-1);
                    UndoMove(i);
                }

                // chance that next move is better is really low.
                // Null window is used for speedup
                alpha = value;
                beta  = alpha+1; 

                UpdateMoveDisplay(&move_stack[i]);
                CopyMainVariant(i);
                moves_in_line = 0;
            }
            // there exists already best move for this depth
            else { 
                if (value>alpha) {
                    // new best move found. calculate its exact value 
                    bestvalue = alpha;
                    alpha     = value;
                    beta      = VALUE_MATT;
                    ExecMove(i);
                    value = -AlphaBetaSearch(-beta,-alpha,distance-1);
                    UndoMove(i);
                    // is it better if we use an open window?
                    if (value>bestvalue) {
                        alpha = value;
                        beta  = alpha+1;
                        UpdateMoveDisplay(&move_stack[i]);
                        CopyMainVariant(i);
                        moves_in_line = 0;
                        // move best move on top 
                        tmp=move_stack[i];
                        for (j=i;j>0;j--) {
                            move_stack[j] = move_stack[j-1];
                        }
                        move_stack[0]=tmp;
                    }
                }
            }
        }
    }

    if (alpha> -(VALUE_MATT-1)) {

        if (act_color == BLACK) {
            sprintf(s,"[%03d] ..........  %s\n",move_count,Move2Str(&move_stack[0]));
            puts(s);
        }
        else {
            sprintf(s,"[%03d] %s\n",move_count,Move2Str(&move_stack[0]));
            puts(s);
        }
        UpdateMoveDisplay(&move_stack[0]);

        ExecMove(0);

        if (alpha>=VALUE_MATT-10) {
            if ((VALUE_MATT-2-alpha)/2 == 0) {
                puts("\n--- that was a clear strike\n");
                puts("--- sorry. try again.\n");
            }
            else {
                sprintf(s,"-> your death is near (%d mvs)\n",(VALUE_MATT-2-alpha)/2);
                puts(s);
            }
        }
        else {
            if(alpha <= -VALUE_MATT+10) {
                sprintf(s,"-> my death is near (%d mvs)\n",(alpha + VALUE_MATT -1)/2);
                puts(s);
            }
        }
        
        return &move_stack[0];
    }
    
    
    if (chess_tmp == TRUE) {
        already_matt = TRUE;
        puts("\n--- Oh god.\n");
        puts("--- I'm such a looser =8-(\n");
    }
    else {
        already_matt = TRUE;
        puts("\n--- A so called PATT\n");
    }

    return 0;
}


/*===========================================================================*/
/* executes an user move                                                     */
/*===========================================================================*/
int UserMove(char* input) {
    int  from,to,i,tmp;

    if(strlen(input)<4) return FALSE;

    from  = Notation2Field(input);
    to    = Notation2Field(&input[2]);
    GenerateMoves(1);

    // loop through all generated moves
    for (i=stack_ranges[act_depth];i<stack_ranges[act_depth+1];i++) {
        //---------------------------------------
        // we've found the corresponding move ...
        //---------------------------------------
        if (move_stack[i].from == from && move_stack[i].to == to) {
            if (move_stack[i].promoted_fig != EMPTY) {
                //-------------------------------------------------------
                // promotions are generated in the following order:
                // queen/knight/bishop/rook -> if we are not promoting to
                // queen we have to increment i !!
                //-------------------------------------------------------
                if (input[4]=='N')     i++;
                else if(input[4]=='B') i+=2;
                else if(input[4]=='R') i+=3;
            }

            InitTree();
            tmp = last_move;
            ExecMove(i); // WARNING: will change color !!!

            //------------------------------------------------------
            // check if our king is still being attacked. if king is
            // under attack -> undo given move
            // NOTE: the reverse checking is necessary, because
            //       ExecMove() has changed the color already
            //------------------------------------------------------
            if (AttacksField((act_color==BLACK)?white_king_pos:black_king_pos,act_color)==TRUE) {
                UndoMove(i);
                last_move=tmp;
                return FALSE;
            }

            return TRUE;
        }
    }

    return FALSE;  // haven't found move in generated move list
}


//##########################################################################
//###################### NO MORE FAKES BEYOND THIS LINE ####################
//##########################################################################

