/******************************************************************************
*
* project name:    TICHESS
* project version: 1.00
* file name:       tichess.c
* file version:    1.04
* initial date:    20/02/2000
* author:          thomas.nussbaumer@peem.com
* description:     just the entry point
*
* -----------------------------------------------------------------------------
* Revision History of TICHESS project
* -----------------------------------------------------------------------------
* v0.91: 22/02/2000
*        finished and distributed (first playable version)
*
* v0.95: 23/02/2000
*        (1)  help page
*        (2)  load from and save to STRING "ticsave"
*        (3)  restart game
*        (4)  change sides
*        (5)  interactive level changing
*        (6)  automatic play mode
*        (7)  show all played moves feature
*        (8)  promotion query
*        (9)  source restructured
*        (10) size reduced to 20kB by using malloc()/free()
*        (11) take back feature
*
* v0.97: 24/02/2000
*        (1) source code preparation for release
*        (2) long running calculation may be cancelled by key press again
*        (3) handling of cursor completely modified
*        (4) TURN board display by 180 degree implemented
*        (5) using now macros for key codes (prepared for porting to TI89)
*
* v0.99: 29/02/2000
*        (1) level-4 computation enabled
*        (2) calculation cancelation will query user
*        (3) engine speedups (now normally near 200 nodes/sec)
*        (4) compiler define for statistics added (DO_STATISTICS)
*        (5) graphical artifacts removed (line height increased)
*        (6) two real player mode added
*
* v1.00: 02/03/2000
*        (1) ported to TI89
*        (2) sourcecode distribution
*        (3) make HW platform define-able
*        (4) splashscreen added
*
*
******************************************************************************/

#include <nostub.h>
#include <all.h>

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// due to a very strange behavior of TIGCC we cannot use separate object until
// yet (the globals are not initialized correctly - NOT even the constants !!)
//
// So we've to go that really silly "include all into one" way ...
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


#include "gui.c"      // gui stuff
#include "engine.c"   // engine stuff

#if TI92_VERSION
int _ti92plus;   // needed to produce ti92plus code
#else 
int _ti89;       // needed to produce ti89 code
#endif

//--------------------------------------------------
// declarations of the necessary "external" routines
//--------------------------------------------------
int  InitMemory();
void FreeMemory();
void MainGUI(int level);


void SplashScreen(void) {
    int i,j;
    int x00,y00,x01,y01;
    int x00_o,y00_o,x01_o,y01_o;
    int maxx    = ScrRect->xy.x1;
    int maxy    = ScrRect->xy.y1;
    int deltax0 = -1;
    int deltay0 = -1;
    char*    s1 = "The TI-Chess Team";
    char*    s2 = "proudly presents";
    char*    s3 = "TI-Chess v1.00";
    
    SetPlane(1);
    ClrScr();
    SetPlane(0);
    ClrScr();

    x00 = maxx/3;
    y00 = maxy/4;
    x01 = maxx*2/3;
    y01 = maxy/2;

    SetPlane(1);
    FontSetSys(F_8x10);
    i = DrawStrWidth(s1,F_8x10); 
    DrawStr((ScrRect->xy.x1-i)/2,ScrRect->xy.y1/4,s1,A_REPLACE);
    FontSetSys(F_6x8);
    i = DrawStrWidth(s2,F_6x8); 
    DrawStr((ScrRect->xy.x1-i)/2,ScrRect->xy.y1/2,s2,A_REPLACE);
    FontSetSys(F_8x10);
    i = DrawStrWidth(s3,F_8x10); 
    DrawStr((ScrRect->xy.x1-i)/2,ScrRect->xy.y1*3/4,s3,A_REPLACE);
    SetPlane(0);

    for (i=0;i<10000;i++) {
        if (kbhit()) {
            ngetchx();
            break;
        }
        DrawLine(x00,y00,x01,y01,A_NORMAL);
        x00_o = x00;
        x01_o = x01;
        y00_o = y00;
        y01_o = y01;
        x00+=deltax0;
        x01+=deltax0;
        y00+=deltay0;
        y01+=deltay0;
        
        if (x00 <= 1 || x01 <= 1 || x00 >= maxx-1 || x01 >= maxx-1) {
           deltax0 = -deltax0;
        }
        if (y00 <= 1 || y01 <= 1 || y00 >= maxy-1 || y01 >= maxy-1) {
           deltay0 = -deltay0;
        }
        
        for (j=0;j<5000;j++);
        DrawLine(x00_o-1,y00_o-1,x01_o-1,y01_o-1,A_XOR);
    }
}


/*===========================================================================*/
/* MAIN - where all the fun starts ...                                       */
/*===========================================================================*/
void _main(void) {
    ESI        argptr;
    int        argtype;
    int        level = 2;

    LCD_BUFFER buffer;

    LCD_save(buffer);
    ClrScr();

    if (InitMemory() == FALSE) {
        FreeMemory();
        DrawStr(0,0,"Insufficient memory (need around 25kB).",A_REPLACE);
        ngetchx();
        LCD_restore(buffer);
        return;
    }

#if HWX_PLATFORM
    DrawStr(0,0,"Hardware release 2 (y/n)?",A_REPLACE);
    if(!GrayMode(((ngetchx()|32)=='y')?GRAY_HW2:GRAY_HW1)) return;
#elif HW1_PLATFORM
    if(!GrayMode(GRAY_HW1)) return;
#elif HW2_PLATFORM
    if(!GrayMode(GRAY_HW2)) return;
#endif

    InitArgPtr(argptr);
    if ((argtype=GetArgType(argptr))!=END_TAG) {
        if (argtype==POSINT_TAG) {
          level=(int)GetIntArg(argptr);
          level++;
          if (level > MAX_USERLEVEL) level = MAX_USERLEVEL;
          if (level < MIN_USERLEVEL) level = MIN_USERLEVEL;
        }
    }

#if DO_STATISTICS
    InitStatistics();
#endif

    SplashScreen();

    MainGUI(level);

    GrayMode(GRAY_OFF);
    FreeMemory();
    LCD_restore(buffer);
}

//##########################################################################
//###################### NO MORE FAKES BEYOND THIS LINE ####################
//##########################################################################
